@@ -27,7 +27,7 @@ import serial
 import serial.tools.list_ports
 from result import Err, Ok, Result
 
-from freewili.types import ButtonColor, EventType, FreeWiliProcessorType, IOMenuCommand
+from freewili.types import ButtonColor, EventType, FileSystemContents, FreeWiliProcessorType, IOMenuCommand
 
 
 @dataclasses.dataclass
@@ -823,6 +823,25 @@ class FreeWiliSerial:
         self.serial_port.send(cmd)
         return self._handle_final_response_frame()
 
+    @needs_open()
+    def enable_audio_events(self, enable: bool) -> Result[str, str]:
+        """Enable or disable audio events.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable audio events.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"a\ns\n{0 if not enable else 1}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
     def process_events(self, delay_sec: float | None = None) -> None:
         """Process events from the FreeWili.
 
@@ -943,12 +962,12 @@ class FreeWiliSerial:
         return self._handle_final_response_frame()
 
     @needs_open()
-    def write_uart(self, data: bytes) -> Result[str, str]:
+    def write_uart(self, data: bytes | str) -> Result[str, str]:
         """Write uart data.
 
         Parameters:
         ----------
-            data : bytes
+            data : bytes | str
                 The data to write.
 
         Returns:
@@ -957,15 +976,136 @@ class FreeWiliSerial:
                 Ok(bytes) if the command was sent successfully, Err(str) if not.
         """
         self._empty_all()
+        assert isinstance(data, (bytes, str)), "data must be bytes or str"
+        if isinstance(data, str):
+            data = data.encode("utf-8")
         data_str = " ".join(f"{b:02x}" for b in data)
         cmd = f"u\nw\n{data_str}"
         self.serial_port.send(cmd)
         return self._handle_final_response_frame()
 
     @needs_open()
-    def enable_stream(self, enable: bool) -> None:
-        """TODO: Docstring."""
-        raise NotImplementedError
+    def send_ir(self, data: bytes) -> Result[str, str]:
+        """Send IR data.
+
+        Notes: v54 firmware uses NEC format and is converted to an 32-bit integer.
+
+        Parameters:
+        ----------
+            data : bytes
+                The data to send. The first 4 bytes are used as the command.
+
+        Returns:
+        -------
+            Result[bytes, str]:
+                Ok(bytes) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        assert isinstance(data, bytes), "data must be bytes"
+        data_int: int = int.from_bytes(data[:4], "big")
+        cmd = f"i\na\n{data_int}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def play_audio_file(self, file_name: str) -> Result[str, str]:
+        """Play an audio file on the FreeWili.
+
+        Arguments:
+        ----------
+        file_name: str
+            Name of the file in the FreeWili. 8.3 filename limit exists as of V12
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"a\nf\n{file_name}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def play_audio_asset(self, asset_value: str | int) -> Result[str, str]:
+        """Play an audio asset on the FreeWili.
+
+        Arguments:
+        ----------
+        asset_value: str | int
+            The asset value to play.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"a\na\n{asset_value}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def play_audio_number_as_speech(self, value: int) -> Result[str, str]:
+        """Play an audio number as speech on the FreeWili.
+
+        Arguments:
+        ----------
+        value: int
+            The audio number to play.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"a\nn\n{value}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def play_audio_tone(self, frequency_hz: int, duration_sec: float, amplitude: float) -> Result[str, str]:
+        """Play an audio tone on the FreeWili.
+
+        Arguments:
+        ----------
+        frequency_hz: int
+            The frequency of the tone in Hertz.
+        duration_sec: float
+            The duration of the tone in seconds.
+        amplitude: float
+            The amplitude of the tone (0.0 to 1.0).
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        assert isinstance(frequency_hz, int)
+        cmd = f"a\nt\n{frequency_hz} {duration_sec:.2f} {amplitude:.2f}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def record_audio(self, file_name: str) -> Result[str, str]:
+        """Record audio on the FreeWili.
+
+        Arguments:
+        ----------
+        file_name: str
+            Name of the file in the FreeWili. (ie. "/sounds/test.wav")
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"a\nr\n{file_name}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
 
     @needs_open()
     def run_script(self, file_name: str) -> Result[str, str]:
@@ -1221,33 +1361,6 @@ class FreeWiliSerial:
                 continue
         return Ok(None) if success else Err("Failed to reset to UF2 bootloader after multiple attempts.")
 
-    def _wait_for_serial_data(self, timeout_sec: float, delay_sec: float = 0.1) -> None:
-        """Wait for data to be available on the serial port.
-
-        Parameters:
-        ----------
-            timeout_sec: float
-                The maximum amount of time to wait for data.
-            delay_sec: float
-                The amount of time to wait after checks for data.
-
-        Returns:
-        -------
-            None
-
-        Raises:
-        -------
-            TimeoutError
-                If the timeout is reached before data is available.
-        """
-        raise NotImplementedError("TODO")
-        # start = time.time()
-        # while self._serial.in_waiting == 0:
-        #     time.sleep(0.001)
-        #     if time.time() - start > timeout_sec:
-        #         raise TimeoutError(f"Timed out waiting for data on {self}")
-        # time.sleep(delay_sec)
-
     @needs_open()
     def get_app_info(self) -> Result[FreeWiliAppInfo, str]:
         """Detect the processor type of the FreeWili.
@@ -1275,47 +1388,245 @@ class FreeWiliSerial:
         else:
             return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Unknown, int(version)))
 
-        self.serial_port.send("", True, "\r\n\r\n")
-        time.sleep(3)
-        all_data = []
-        while True:
+    @needs_open()
+    def change_directory(self, directory: str) -> Result[str, str]:
+        """Change the current directory on the FreeWili.
+
+        Arguments:
+        ----------
+            directory: str
+                The directory to change to.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"x\na\n{directory}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def create_directory(self, directory: str) -> Result[str, str]:
+        """Create a new directory on the FreeWili.
+
+        Arguments:
+        ----------
+            directory: str
+                The directory to create.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"x\nc\n{directory}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def remove_directory_or_file(self, dir_or_filename: str) -> Result[str, str]:
+        """Remove a directory or file on the FreeWili.
+
+        Arguments:
+        ----------
+            dir_or_filename: str
+                The directory or file to remove.
+                The directory to create.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"x\nr\n{dir_or_filename}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def create_blank_file(self, name: str) -> Result[str, str]:
+        """Create a blank file on the FreeWili.
+
+        Arguments:
+        ----------
+            name: str
+                The name of the file to create.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"x\nb\n{name}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def move_directory_or_file(self, original_name: str, new_name: str) -> Result[str, str]:
+        """Move a directory or file on the FreeWili.
+
+        Arguments:
+        ----------
+            original_name: str
+                The original name of the directory or file to move.
+            new_name: str
+                The new name of the directory or file.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = f"x\nn\n{original_name} {new_name}"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open()
+    def format_filesystem(self) -> Result[str, str]:
+        """Format the filesystem on the FreeWili.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        self._empty_all()
+        cmd = "x\nt\ndestroyfiles"
+        self.serial_port.send(cmd)
+        return self._handle_final_response_frame()
+
+    @needs_open(enable_menu=True)
+    def list_current_directory(self) -> Result[FileSystemContents, str]:
+        """List the contents of the current directory on the FreeWili.
+
+        Returns:
+        -------
+            Result[FileSystemContents, str]:
+                Ok(FileSystemContents) if the command was sent successfully, Err(str) if not.
+        """
+        from freewili.types import FileSystemContents, FileSystemItem, FileType
+
+        self._empty_all()
+        # re-print the menu to get the filesystem listing
+        cmd = ""
+        self.serial_port.send(cmd)
+
+        # Collect all output data
+        collected_data = ""
+        start = time.time()
+        while time.time() - start <= 6.0:
             try:
-                data = self.serial_port.data_queue.get_nowait()
-                all_data.append(data)
-            except Empty:
+                data = self.serial_port.data_queue.get_nowait().decode("utf-8", errors="ignore")
+                collected_data += data
+                if "Enter Letter:" in data:
+                    break
+            except queue.Empty:
+                time.sleep(0.001)
+                continue
+
+        # Remove ANSI escape codes and normalize line endings
+        import re
+
+        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
+        collected_data = ansi_escape.sub("", collected_data)
+        collected_data = collected_data.replace("\r\n", "\n").replace("\r", "\n")
+
+        # Parse the filesystem output
+        lines = [line.strip() for line in collected_data.split("\n") if line.strip()]
+
+        # Find the filesystem section
+        fs_start_idx = None
+        cwd = "/"
+
+        for i, line in enumerate(lines):
+            if "File System" in line and "bytes free" in line:
+                fs_start_idx = i
                 break
 
-        self._wait_for_serial_data(3.0)
-        data = self._serial.read_all()
-        # proc_type_regex = re.compile(r"(Main|Display) Processor")
-        # match = proc_type_regex.search(data.decode())
-        # if match is None:
-        #     return Ok(FreeWiliProcessorType.Unknown)
-        # elif "Main Processor" in match.group():
-        #     return Ok(FreeWiliProcessorType.Main)
-        # elif "Display Processor" in match.group():
-        #     return Ok(FreeWiliProcessorType.Display)
-        # else:
-        #     return Err("Unknown processor type detected!")
-        line = ""
-        for line in data.decode().splitlines():
-            if "Processor" in line or "MainCPU" in line or "DisplayCPU" in line:
+        if fs_start_idx is None:
+            return Err("Could not find filesystem information in output")
+
+        # Look for directory line (e.g., "/ directory contents")
+        for i in range(fs_start_idx + 1, len(lines)):
+            line = lines[i].strip()
+            if "directory contents" in line:
+                cwd = line.split(" directory contents")[0].strip()
+                fs_start_idx = i
                 break
-        proc_type_regex = re.compile(r"(?:Main|Display)|(?:App version)|(?:\d+)")
-        results = proc_type_regex.findall(line)
-        if len(results) == 2:
-            # New firmware >= 48
-            processor = results[0]
-            version = results[1]
-        elif len(results) == 3:
-            # Legacy firmware
-            processor = results[0]
-            version = results[2]
-        else:
-            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Unknown, 0))
-        if "Main" in processor:
-            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Main, int(version)))
-        elif "Display" in processor:
-            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Display, int(version)))
-        else:
-            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Unknown, 0))
+
+        if fs_start_idx is None:
+            return Err("Could not find directory contents section")
+
+        # Parse file/directory entries
+        items = []
+        for i in range(fs_start_idx + 1, len(lines)):
+            line = lines[i].strip()
+            if not line:
+                continue
+
+            # Skip menu lines or other non-filesystem content
+            menu_prefixes = (
+                "a)",
+                "b)",
+                "c)",
+                "d)",
+                "e)",
+                "f)",
+                "g)",
+                "h)",
+                "i)",
+                "j)",
+                "k)",
+                "l)",
+                "m)",
+                "n)",
+                "o)",
+                "p)",
+                "q)",
+                "r)",
+                "s)",
+                "t)",
+                "u)",
+                "v)",
+                "w)",
+                "x)",
+                "y)",
+                "z)",
+            )
+            if line.startswith(menu_prefixes):
+                break
+
+            # Parse file/dir entries: "file  settings.txt  0 bytes" or "dir   scripts"
+            parts = line.split()
+            if len(parts) >= 2:
+                file_type_str = parts[0]
+                name = parts[1]
+
+                if file_type_str == "file" and len(parts) >= 4 and parts[3] == "bytes":
+                    # File entry: "file  settings.txt  0 bytes"
+                    try:
+                        size = int(parts[2])
+                        items.append(FileSystemItem(name=name, file_type=FileType.File, size=size))
+                    except ValueError:
+                        # If size parsing fails, default to 0
+                        items.append(FileSystemItem(name=name, file_type=FileType.File, size=0))
+                elif file_type_str == "dir":
+                    # Directory entry: "dir   scripts"
+                    items.append(FileSystemItem(name=name, file_type=FileType.Directory, size=0))
+
+        return Ok(FileSystemContents(cwd=cwd, contents=items))
+
+
+"""
+File System (14381056 bytes free of 14393344)
+/ directory contents
+file  settings.txt  0 bytes
+dir   scripts
+dir   data
+dir   radio
+"""
